# re-write of 'facegroup.rb' (version 0.1)
# version 0.5.1
# 2011-01-12
#
#	Revision Log for v0.5.1
#	- added Jim's version checking for start_operation (thanks to Jim and Daarboven)
#
# Revision Log for v0.5
#
# - refactored code around new operations
# - changed menu style (added edge settings)
# - removed iteration issue by revising commons instead of growing faces (sketchup crashes at 85th iteration of definition)
# - replaced refined pass technique with references update to sortCommonFaces (huge speed increase)
#
# re-write by bentleykf(niall.campbell@gmail.com) (from version 0.2 onwards)
# 'facegroup.rb' written by FF(gavvy@163.com)
# 'facegroup.rb' is a 90% copy from 'PUSH_PULL_Z_SET.rb'
# 'PUSH_PULL_Z_SET.rb' by D. Bur (http://rhin.crai.archi.fr/RubyLibraryDepot/)
###
require 'sketchup.rb'
#Sketchup.send_action "showRubyPanel:"
###
module FaceGroup

  #############################################
  # this definition converts the selected face(face) into a group
  def self.faceToGroup(faces)
      
       begin
         # lets try to create a group then copy entities to it,
         # then erase the entities in the faces definition after completion
        Sketchup.active_model.active_entities.add_group(faces)
       rescue
         return nil
         #do something with spare geometry?
       end



    #end #if
    return nil

  end #def
  #############################################
	
	def self.getObjectsFromArray(entities_array,object_type,msg = nil)
		return_hash = Hash.new()
		return return_hash if (!entities_array.kind_of?(Array) || entities_array.length <= 0)

		tick=0
		msg_status = msg
		entities_length = entities_array.length.to_s
		case object_type
			when "faces"
				entities_array.each{|entity|
					msg_status=msg+": "+tick.to_s+"/"+entities_length
					Sketchup::set_status_text(msg_status,SB_PROMPT)
					return_hash[entity.entityID] = entity if entity.kind_of?(Sketchup::Face)
					tick+=1
				}
			when "edges"
				entities_array.each{|entity|
					msg_status=msg+": "+tick.to_s+"/"+entities_length
					Sketchup::set_status_text(msg_status,SB_PROMPT)
					if entity.kind_of?(Sketchup::Face)
						entity.edges.each{|edge|
							return_hash[edge.entityID] = edge
						}
					end #if
					tick+=1
				}
		end #case
		return return_hash
	end #def
	
	
	def self.unsoftenEdgesFromHash(entities_hash,entity_type)
		edge_treatment = 1
		edge_treatment = 1 if Sketchup.read_default("Faces To Groups","RHE",true)
		edge_treatment = 2 if Sketchup.read_default("Faces To Groups","PEC",true)
		edge_treatment = 3 if Sketchup.read_default("Faces To Groups","AEH",true)
		
		case edge_treatment
			when 1 #restore the hidden edges (ie unhide)
				case entity_type
					when "faces"
						entities_hash.each{|face_id, f| f.edges.each{|e|
							if (e.faces.length == 1) then e.soft = false; e.smooth = false; e.hidden = false; end #make outside edges hard
							e.hidden = false
						}}
					when "edges"
						entities_hash.each{|edge_id, e| 
							if (e.faces.length == 1) then e.soft = false; e.smooth = false; e.hidden = false; end #make outside edges hard
							e.hidden = false
						}
				end #case
			when 2 #preserve the edges condition (ie don't change the edge)
				#do nothing here :)
			when 3 #turn all edges hard
				case entity_type
					when "faces"
						entities_hash.each{|face_id, f| f.edges.each{|e| e.soft = false; e.smooth = false; e.hidden = false}}
					when "edges"
						entities_hash.each{|edge_id, e| e.soft = false; e.smooth = false; e.hidden = false}
				end #case
		end #case
	end #def

	
	
	def self.getCommonFaces(edges_hash,sort_type,msg = nil)
		unsorted_common = {"faces" => Hash.new(),"edges" => Hash.new(),"objects" => Hash.new()}
		edge_treatment = 1
		edge_treatment = 1 if Sketchup.read_default("Faces To Groups","RHE",true)
		edge_treatment = 2 if Sketchup.read_default("Faces To Groups","PEC",true)
		edge_treatment = 3 if Sketchup.read_default("Faces To Groups","AEH",true)

		
		
		tick = 0
		msg_status = msg
		edges_length = edges_hash.length.to_s
		edges_hash.each{|edge_id, edge|                                              #go through each of the edges
			msg_status=msg+": "+tick.to_s+"/"+edges_length
			Sketchup::set_status_text(msg_status,SB_PROMPT)

			case sort_type
				when 3
					edge_condition = (edge.soft? || edge.smooth?)
				when 4
					edge_condition = (edge.soft? || edge.smooth? || edge.hidden?)
			end #case

			if edge_condition || (edge.faces.length == 1)
				case edge_treatment
					when 1 #restore the hidden edges (ie unhide)
						if (edge_condition && edge.faces.length == 1) then edge.soft = false; edge.smooth = false; edge.hidden = false; end #make outside edge hard
						edge.hidden = false #unhide the hidden edges
					when 2 #preserve the edges condition (ie don't change the edge)
						#do nothing here :)
					when 3 #turn all edges hard
						edge.soft   = false
						edge.smooth = false
						edge.hidden = false
				end #case

				edge.faces.each{|face|																										 #and find the common faces
					if !unsorted_common["faces"][edge_id]		                       #store the associations in arrays to be sorted
						unsorted_common["faces"][edge_id] = Array[face.entityID]			 #NOTE: enter in edge condition!!!
						unsorted_common["objects"][edge_id] = Array[face]
					else
						unsorted_common["faces"][edge_id] << face.entityID
						unsorted_common["objects"][edge_id] << face
					end #if
					if !unsorted_common["edges"][face.entityID]
						unsorted_common["edges"][face.entityID] = Array[edge_id]
					else
						unsorted_common["edges"][face.entityID] << edge_id
					end #if
				}
			else
				if(edge.hidden?)
					case edge_treatment
						when 1 #restore the hidden edges (ie unhide)
							if (edge.faces.length == 1) then edge.soft = false; edge.smooth = false; edge.hidden = false; end #make outside edge hard
							edge.hidden = false #unhide the hidden edges
						when 2 #preserve the edges condition (ie don't change the edge)
							#do nothing here :)
						when 3 #turn all edges hard
							edge.soft   = false
							edge.smooth = false
							edge.hidden = false
					end #case
					#need to run something to group each remaining faces together
				end #if
			end #if
			tick += 1
		}

		return unsorted_common
	end #def
	
	def self.sortCommonFaces(unsorted_edges,unsorted_faces,unsorted_objects,msg)
		#so unsorted_edges contains the common edges to a face    x[face_id] = [edge_id,edge_id,etc]
		#and unsorted_faces contains the common faces to an edge  x[edge_id] = [face_id,face_id,etc]

		pass = 1
		tick = 0
		commons = 0
		msg_status = msg
		find_more_similarities = true
		revised_faces = Array.new() #--------------
		while find_more_similarities do															#refactoring loop, will be untrue when no similarities found
			tick = 0
			commons = 0
			edges_length = unsorted_edges.length.to_s
			find_more_similarities = false														#reset the while switch to run the check
			unsorted_edges.each{|face_id, edge_id_array|              #loop through the unsorted edge reference arrays
				msg_status=msg+":: Pass:"+pass.to_s+" Blocks:"+tick.to_s+"/"+edges_length+" Commons Found:"+commons.to_s
				Sketchup::set_status_text(msg_status,SB_PROMPT)
				revised_faces.clear																		  #clear the revised faces array----------
				edge_id_array.each{|edge_id|														#loop through the common edges
					unsorted_faces[edge_id].each{|cface_id|										#remove the reference so the check doesn't complete next time
						if (unsorted_edges[cface_id] && !(cface_id == face_id))
							find_more_similarities = true
							unsorted_edges[face_id] = unsorted_edges[face_id] | unsorted_edges[cface_id]	#combine references together
							unsorted_edges.delete(cface_id)                         #remove the edge_id_array from unsorted edges
							commons+=1
						end #if																								    #as its been checked							
					}
					revised_faces = revised_faces | unsorted_faces[edge_id]	  #create revised data for unsorted_faces[edge_id] arrays--------------
				}
				edge_id_array.each{|edge_id|																#revise reference arrays-------------
					unsorted_faces[edge_id] = revised_faces.dup
				}
				tick+=1
			}
      #puts "Pass:"+pass.to_s+" Commons:"+commons.to_s
			pass+=1
		end #while

		# refined pass, cycles through remaining arrays and if common joins arrays
		# NB: Following code removed after adding reference array update code (unsorted_faces[edge_id] = revised_faces.dup)
		# With    ref array update code: Small Op: 0.618s, Medium Op: 1.570s, Large Op:  5.385s, Massive Op:  17.064s
		# Without ref array update code: Small Op: 0.530s, Medium Op: 2.401s, Large Op: 14.085s, Massive Op: 286.707s
		#find_more_similarities = true
		
		#while find_more_similarities do
		#	tick = 0
		#	commons = 0
		#	edges_length = unsorted_edges.length.to_s
		#	find_more_similarities = false
		#	unsorted_edges.each{|face_id, edge_id_array|
		#		msg_status=msg+":: Refined Pass:"+pass.to_s+" Blocks:"+tick.to_s+"/"+edges_length+" Commons Found:"+commons.to_s
		#		Sketchup::set_status_text(msg_status,SB_PROMPT)
		#		edge_id_array.each{|edge_id|
		#			unsorted_edges.each{|cface_id, cedge_id_array|
		#				next if face_id == cface_id
		#				cedge_id_array.each{|cedge_id|
		#					next if edge_id != cedge_id
		#					unsorted_edges[face_id] = unsorted_edges[face_id] | unsorted_edges[cface_id]
		#					unsorted_edges.delete(cface_id)
		#					find_more_similarities = true
		#					commons +=1
		#				}
		#			}
		#		}
		#		tick+=1
		#	}
    #  puts "Refined Pass:"+pass.to_s+" Commons:"+commons.to_s
		#	pass+=1
		#end #while

		

		msg = "Combining Common Face(s)"
		#whats left in unsorted_edges is edge_id group references to unsorted_objects
		#these edge_ids will be used to combine hashes into each group
		sorted_objects = Hash.new()
		pass=1
		unsorted_edges.each{|face_id, edge_id_array|
			tick=0
			edges_length = edge_id_array.length.to_s
			edge_id_array.each{|edge_id|
				msg_status=msg+":: Group:"+pass.to_s+" Faces:"+tick.to_s+"/"+edges_length
				Sketchup::set_status_text(msg_status,SB_PROMPT)
				if !sorted_objects[face_id]
					sorted_objects[face_id] = unsorted_objects[edge_id]
				else
					sorted_objects[face_id] = sorted_objects[face_id] | unsorted_objects[edge_id]
				end #if
				tick+=1
			}
			pass+=1
			
		}
		
		return sorted_objects
		
	end #def
	
	def self.sortEntities(entities_array,sort_type)
		case sort_type
			when 1 #group by any edge and unsoften
				msg="Gathering Face(s)"
				faces_hash = FaceGroup.getObjectsFromArray(entities_array,"faces",msg)
				FaceGroup.unsoftenEdgesFromHash(faces_hash,"faces")
				return faces_hash
			when 3 #group by hard edge
				msg="Gathering Edge(s)"
				edges_hash = FaceGroup.getObjectsFromArray(entities_array,"edges",msg)
				msg="Gathering Face(s)"
				common_entities = FaceGroup.getCommonFaces(edges_hash,sort_type,msg)
				msg="Finding Common Face(s)"
				faces_hash = FaceGroup.sortCommonFaces(common_entities["edges"],common_entities["faces"],common_entities["objects"],msg)
				return faces_hash
			when 4 #group by hard edge and hidden = soft
				msg="Gathering Edge(s)"
				edges_hash = FaceGroup.getObjectsFromArray(entities_array,"edges",msg)
				msg="Gathering Face(s)"
				common_entities = FaceGroup.getCommonFaces(edges_hash,sort_type,msg)
				msg="Finding Common Face(s)"
				faces_hash = FaceGroup.sortCommonFaces(common_entities["edges"],common_entities["faces"],common_entities["objects"],msg)
				return faces_hash
		end #case
	end #def


  #############################################
  # this definition finds the selected faces from selected entities
	# (sortEntities) and then cycles through selected faces(face_groups_hash) 
  # and runs faceToGroup for each possible face

  def self.facesToGroup(entities_array,sort_type)
    model=Sketchup.active_model
		#run Jim's sketchup version check
		if Sketchup.version.to_f < 7.0
			model.start_operation("Face(s) to Group(s)",true)
		else
			model.start_operation("Face(s) to Group(s)")
		end #if
		#t=Time.now
    msg="Grouping Face(s)"
    #try and sort out faces from the entities array
    face_groups_hash = FaceGroup.sortEntities(entities_array,sort_type)

	  if face_groups_hash.length <= 0
	    UI.messagebox "No faces found in selection", MB_MULTILINE
	    Sketchup::set_status_text("Operation Aborted",SB_PROMPT)
	    model.abort_operation
	    return nil
	  end #if

	  if face_groups_hash.length > 500
	    result = UI.messagebox "The groups created from the faces selected will exceed 500. This will dramatically slow down Sketchup. Click 'Yes' to continue script. (Total Groups: "+face_groups_hash.length.to_s+")", MB_YESNO
	    if result == 7 #no
	      Sketchup::set_status_text("Operation Aborted",SB_PROMPT)
	      model.abort_operation
	      return nil
	    end #if
	  end #if
		
    tick=0
		msg_status=""
		faces_length = face_groups_hash.length.to_s

		face_groups_hash.each{|key, faces|
			msg_status=msg+": "+tick.to_s+"/"+faces_length
			Sketchup::set_status_text(msg_status,SB_PROMPT)
			FaceGroup.faceToGroup(faces)
			tick+=1
		}
    Sketchup::set_status_text("Operation Complete: "+tick.to_s+" face groups",SB_PROMPT)
    model.commit_operation
		#puts "Total Time: "+(Time.now - t).to_s
    return nil

  end #def
  #############################################

end #module

#############################################
# this part adds the plugin option to the plugin menu
if not file_loaded?(File.basename(__FILE__))
  submenu = UI.menu("Plugins").add_submenu("Face(s) to Group(s)")
  submenu.add_item("F2G: by Any Edge"){
    FaceGroup.facesToGroup(Sketchup.active_model.selection.to_a,1)
    Sketchup.active_model.selection.clear
  }
  submenu.add_item("F2G: by Hard Edge"){
    FaceGroup.facesToGroup(Sketchup.active_model.selection.to_a,4)
    Sketchup.active_model.selection.clear
  }
  submenu.add_item("F2G: by Hard + Hard Hidden"){
    FaceGroup.facesToGroup(Sketchup.active_model.selection.to_a,3)
    Sketchup.active_model.selection.clear
  }
	submenu.add_separator
	
	#set the defaults if they haven't already been set
	Sketchup.write_default("Faces To Groups","PEC",true) if Sketchup.read_default("Faces To Groups","PEC","fail") == "fail"
	Sketchup.write_default("Faces To Groups","RHE",false) if Sketchup.read_default("Faces To Groups","RHE","fail") == "fail"
	Sketchup.write_default("Faces To Groups","AEH",false) if Sketchup.read_default("Faces To Groups","AEH","fail") == "fail"

	cmd = UI::Command.new("Preserve Edge Condition") {
		Sketchup.write_default("Faces To Groups","PEC",true)
		Sketchup.write_default("Faces To Groups","RHE",false)
		Sketchup.write_default("Faces To Groups","AEH",false)
	}
	
	cmd.set_validation_proc {
		Sketchup.read_default("Faces To Groups","PEC",true) ? MF_CHECKED : MF_UNCHECKED
	}
	
	submenu.add_item(cmd)
	
	cmd = UI::Command.new("Restore Hidden Edges") {
		Sketchup.write_default("Faces To Groups","PEC",false)
		Sketchup.write_default("Faces To Groups","RHE",true)
		Sketchup.write_default("Faces To Groups","AEH",false)
	}
	cmd.set_validation_proc {
		Sketchup.read_default("Faces To Groups","RHE",true) ? MF_CHECKED : MF_UNCHECKED
	}
	submenu.add_item(cmd)
	
	cmd = UI::Command.new("All Edges Hard") {
		Sketchup.write_default("Faces To Groups","PEC",false)
		Sketchup.write_default("Faces To Groups","RHE",false)
		Sketchup.write_default("Faces To Groups","AEH",true)
	}
	cmd.set_validation_proc {
		Sketchup.read_default("Faces To Groups","AEH",true) ? MF_CHECKED : MF_UNCHECKED
	}
	submenu.add_item(cmd)
	
	
  UI.add_context_menu_handler do |context_menu|
    face_selected = false
    selected = Sketchup.active_model.selection.to_a
		if(selected[0])
			selected.each{|e|
				if e.kind_of?(Sketchup::Face)
					face_selected = true
					break
				end #if
			}
		end #if
    if face_selected
	  context_menu.add_separator
	  context_submenu = context_menu.add_submenu("Face(s) to Group(s)")
      context_submenu.add_item("F2G: by Any Edge"){
				FaceGroup.facesToGroup(Sketchup.active_model.selection.to_a,1)
        Sketchup.active_model.selection.clear
      }
			context_submenu.add_item("F2G: by Hard Edge"){
        FaceGroup.facesToGroup(Sketchup.active_model.selection.to_a,4)
        Sketchup.active_model.selection.clear
      }
      context_submenu.add_item("F2G: by Hard + Hard Hidden"){
        FaceGroup.facesToGroup(Sketchup.active_model.selection.to_a,3)
        Sketchup.active_model.selection.clear
      }
		end #if
  end #do
end #if
file_loaded(File.basename(__FILE__))
#############################################